import re
import logging
from typing import Optional
from functools import lru_cache
from normality import collapse_spaces, ascii_text, category_replace

from fingerprints.constants import WS, BRACKETED

log = logging.getLogger(__name__)

CHARACTERS_REMOVE_RE = re.compile(r"[\.\'’]")

PREFIXES_RAW_LIST = [
    "Mr",
    "Ms",
    "Mrs",
    "Mister",
    "Miss",
    "Madam",
    "Madame",
    "Monsieur",
    "Mme",
    "Mmme",
    "Herr",
    "Hr",
    "Frau",
    "Fr",
    "The",
    "Fräulein",
    "Senor",
    "Senorita",
    "Sheik",
    "Sheikh",
    "Shaikh",
    "Sr",
    "Sir",
    "Lady",
    "The",
    "de",
    "of",
]
PREFIXES_RAW = "|".join(PREFIXES_RAW_LIST)
NAME_PATTERN_ = r"^\W*((%s)\.?\s+)*(?P<term>.*?)([\'’]s)?\W*$"
NAME_PATTERN_ = NAME_PATTERN_ % PREFIXES_RAW
PREFIXES = re.compile(NAME_PATTERN_, re.I | re.U)


def clean_entity_prefix(name: str) -> str:
    """Remove prefixes like Mr., Mrs., etc."""
    match = PREFIXES.match(name)
    if match is not None:
        name = match.group("term")
    return name


def clean_brackets(text: str) -> str:
    """Remove any text in brackets. This is meant to handle names of companies
    which include the jurisdiction, like: Turtle Management (Seychelles) Ltd."""
    return BRACKETED.sub(WS, text)


@lru_cache(maxsize=2000)
def clean_name_ascii(text: Optional[str]) -> Optional[str]:
    """Super-hardcore string scrubbing."""
    # transliterate to ascii
    text = ascii_text(text)
    if text is None:
        return None
    # replace punctuation and symbols
    text = CHARACTERS_REMOVE_RE.sub("", text)
    text = text.lower()
    cleaned = category_replace(text)
    cleaned = collapse_spaces(cleaned)
    if cleaned is None or len(cleaned) < 2:
        return None
    return cleaned


@lru_cache(maxsize=2000)
def clean_name_light(text: str) -> Optional[str]:
    """Clean up a name for comparison, but don't convert to ASCII/Latin."""
    # replace punctuation and symbols
    text = CHARACTERS_REMOVE_RE.sub("", text)
    text = text.lower()
    cleaned = category_replace(text)
    cleaned = collapse_spaces(cleaned)
    if cleaned is None or len(cleaned) < 2:
        return None
    return cleaned



import re
import logging
from functools import lru_cache
from typing import Callable, Optional, Dict, Match

from fingerprints.types.data import TYPES
from fingerprints.constants import WS
from fingerprints.cleanup import clean_name_ascii

log = logging.getLogger(__name__)
NormFunc = Callable[[Optional[str]], Optional[str]]
ReplaceFunc = Callable[[Optional[str]], Optional[str]]


class Replacer(object):
    def __init__(self, replacements: Dict[str, str], remove: bool = False) -> None:
        self.replacements = replacements
        self.remove = remove
        forms = set(self.replacements.keys())
        if remove:
            forms.update(self.replacements.values())
        forms_sorted = sorted(forms, key=lambda ct: -1 * len(ct))
        forms_regex = "\\b(%s)\\b" % "|".join(forms_sorted)
        self.matcher = re.compile(forms_regex, re.U)

    def get_canonical(self, match: Match[str]) -> str:
        if self.remove:
            return WS
        return self.replacements.get(match.group(1), match.group(1))

    def __call__(self, text: Optional[str]) -> Optional[str]:
        if text is None:
            return None
        return self.matcher.sub(self.get_canonical, text)


def normalize_replacements(norm_func: NormFunc) -> Dict[str, str]:
    replacements: Dict[str, str] = {}
    for type in TYPES["types"]:
        main_norm = norm_func(type["main"])
        if main_norm is None:
            log.warning("Main form is normalized to null: %r", type["main"])
            continue
        for form in type["forms"]:
            form_norm = norm_func(form)
            if form_norm is None:
                log.warning("Form is normalized to null [%r]: %r", type["main"], form)
                continue
            if form_norm == main_norm:
                continue
            if form_norm in replacements and replacements[form_norm] != main_norm:
                log.warning(
                    "Form has duplicate mains: %r (%r, %r)",
                    form,
                    replacements[form_norm],
                    main_norm,
                )
                continue
            replacements[form_norm] = main_norm
    return replacements


@lru_cache(maxsize=None)
def get_replacer(
    clean: NormFunc = clean_name_ascii, remove: bool = False
) -> ReplaceFunc:
    replacements = normalize_replacements(clean)
    return Replacer(replacements, remove=remove)


if __name__ == "__main__":
    get_replacer()



from fingerprints.fingerprint import fingerprint
from fingerprints.cleanup import clean_entity_prefix
from fingerprints.cleanup import clean_brackets
from fingerprints.cleanup import clean_name_light
from fingerprints.cleanup import clean_name_ascii
from fingerprints.types import remove_types, replace_types

generate = fingerprint

__all__ = [
    "fingerprint",
    "generate",
    "clean_entity_prefix",
    "clean_brackets",
    "clean_name_light",
    "clean_name_ascii",
    "remove_types",
    "replace_types",
]
